@@ -15,7 +15,9 @@
 #include <signal.h>
 #include <string.h>
 #include <errno.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif /* HAVE_UNISTD_H */
 #include <sys/socket.h>
 #include <sys/time.h>
 #ifdef HAVE_SYS_STAT_H
@@ -26,6 +28,8 @@
 #include <arpa/inet.h>
 #include <setjmp.h>
 #include <time.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 
 #include <atalk/logger.h>
 #include <atalk/dsi.h>
@@ -35,8 +39,10 @@
 #include <atalk/paths.h>
 #include <atalk/server_ipc.h>
 #include <atalk/fce_api.h>
-
 #include <atalk/globals.h>
+#include <atalk/netatalk_conf.h>
+#include <atalk/spotlight.h>
+
 #include "switch.h"
 #include "auth.h"
 #include "fork.h"
@@ -46,7 +52,7 @@
 #define SOL_TCP IPPROTO_TCP
 #endif
 
-/* 
+/*
  * We generally pass this from afp_over_dsi to all afp_* funcs, so it should already be
  * available everywhere. Unfortunately some funcs (eg acltoownermode) need acces to it
  * but are deeply nested in the function chain with the caller already without acces to it.
@@ -72,25 +78,25 @@ static rc_elem_t replaycache[REPLAYCACHE_SIZE];
 static sigjmp_buf recon_jmp;
 static void afp_dsi_close(AFPObj *obj)
 {
-    DSI *dsi = obj->handle;
+    DSI *dsi = obj->dsi;
     sigset_t sigs;
-    
+
     close(obj->ipc_fd);
     obj->ipc_fd = -1;
 
     /* we may have been called from a signal handler caught when afpd was running
-     * as uid 0, so restore our login user
+     * as uid 0, that's the wrong user for volume's prexec_close scripts if any,
+     * restore our login user
      */
     if (geteuid() != obj->uid) {
         if (seteuid( obj->uid ) < 0) {
-            LOG(log_error, logtype_afpd, "can't seteuid(%u) back %s: uid: %u, euid: %u", 
+            LOG(log_error, logtype_afpd, "can't seteuid(%u) back %s: uid: %u, euid: %u",
                 obj->uid, strerror(errno), getuid(), geteuid());
             exit(EXITERR_SYS);
         }
     }
 
-    close_all_vol();
-
+    close_all_vol(obj);
     if (obj->logout) {
         /* Block sigs, PAM/systemd/whoever might send us a SIG??? in (*obj->logout)() -> pam_close_session() */
         sigfillset(&sigs);
@@ -107,11 +113,11 @@ static void afp_dsi_close(AFPObj *obj)
 
 /* -------------------------------
  * SIGTERM
- * a little bit of code duplication. 
+ * a little bit of code duplication.
  */
 static void afp_dsi_die(int sig)
 {
-    DSI *dsi = (DSI *)AFPobj->handle;
+    DSI *dsi = (DSI *)AFPobj->dsi;
 
     if (dsi->flags & DSI_RECONINPROG) {
         /* Primary reconnect succeeded, got SIGTERM from afpd parent */
@@ -124,7 +130,7 @@ static void afp_dsi_die(int sig)
         exit(0);
     }
 
-    dsi_attention(AFPobj->handle, AFPATTN_SHUTDOWN);
+    dsi_attention(AFPobj->dsi, AFPATTN_SHUTDOWN);
     afp_dsi_close(AFPobj);
    if (sig) /* if no signal, assume dieing because logins are disabled &
                 don't log it (maintenance mode)*/
@@ -137,29 +143,12 @@ static void afp_dsi_die(int sig)
     }
 }
 
-/* SIGQUIT handler */
-static void ipc_reconnect_handler(int sig _U_)
-{
-    if (reconnect_ipc(AFPobj) != 0) {
-        LOG(log_error, logtype_afpd, "ipc_reconnect_handler: failed IPC reconnect");
-        afp_dsi_close(AFPobj);
-        exit(EXITERR_SYS);        
-    }
-
-    if (ipc_child_write(AFPobj->ipc_fd, IPC_GETSESSION, AFPobj->sinfo.clientid_len, AFPobj->sinfo.clientid) != 0) {
-        LOG(log_error, logtype_afpd, "ipc_reconnect_handler: failed IPC ID resend");
-        afp_dsi_close(AFPobj);
-        exit(EXITERR_SYS);        
-    }
-    LOG(log_note, logtype_afpd, "ipc_reconnect_handler: IPC reconnect done");
-}
-
 /* SIGURG handler (primary reconnect) */
 static void afp_dsi_transfer_session(int sig _U_)
 {
     uint16_t dsiID;
     int socket;
-    DSI *dsi = (DSI *)AFPobj->handle;
+    DSI *dsi = (DSI *)AFPobj->dsi;
 
     LOG(log_debug, logtype_afpd, "afp_dsi_transfer_session: got SIGURG, trying to receive session");
 
@@ -197,7 +186,7 @@ static void afp_dsi_transfer_session(int sig _U_)
     }
 
     LOG(log_note, logtype_afpd, "afp_dsi_transfer_session: successful primary reconnect");
-    /* 
+    /*
      * Now returning from this signal handler return to dsi_receive which should start
      * reading/continuing from the connected socket that was passed via the parent from
      * another session. The parent will terminate that session.
@@ -210,15 +199,15 @@ static void afp_dsi_timedown(int sig _U_)
 {
     struct sigaction	sv;
     struct itimerval	it;
-    DSI                 *dsi = (DSI *)AFPobj->handle;
+    DSI                 *dsi = (DSI *)AFPobj->dsi;
     dsi->flags |= DSI_DIE;
     /* shutdown and don't reconnect. server going down in 5 minutes. */
     setmessage("The server is going down for maintenance.");
-    if (dsi_attention(AFPobj->handle, AFPATTN_SHUTDOWN | AFPATTN_NORECONNECT |
+    if (dsi_attention(AFPobj->dsi, AFPATTN_SHUTDOWN | AFPATTN_NORECONNECT |
                   AFPATTN_MESG | AFPATTN_TIME(5)) < 0) {
-        DSI *dsi = (DSI *)AFPobj->handle;
+        DSI *dsi = (DSI *)AFPobj->dsi;
         dsi->down_request = 1;
-    }                  
+    }
 
     it.it_interval.tv_sec = 0;
     it.it_interval.tv_usec = 0;
@@ -273,17 +262,17 @@ static void afp_dsi_debug(int sig _U_)
 /* ---------------------- */
 static void afp_dsi_getmesg (int sig _U_)
 {
-    DSI *dsi = (DSI *)AFPobj->handle;
+    DSI *dsi = (DSI *)AFPobj->dsi;
 
     dsi->msg_request = 1;
-    if (dsi_attention(AFPobj->handle, AFPATTN_MESG | AFPATTN_TIME(5)) < 0)
+    if (dsi_attention(AFPobj->dsi, AFPATTN_MESG | AFPATTN_TIME(5)) < 0)
         dsi->msg_request = 2;
 }
 
 static void alarm_handler(int sig _U_)
 {
     int err;
-    DSI *dsi = (DSI *)AFPobj->handle;
+    DSI *dsi = (DSI *)AFPobj->dsi;
 
     /* we have to restart the timer because some libraries may use alarm() */
     setitimer(ITIMER_REAL, &dsi->timer, NULL);
@@ -313,7 +302,7 @@ static void alarm_handler(int sig _U_)
             afp_dsi_die(EXITERR_CLNT);
         }
         return;
-    } 
+    }
 
     if (dsi->flags & DSI_DISCONNECTED) {
         if (geteuid() == 0) {
@@ -327,7 +316,7 @@ static void alarm_handler(int sig _U_)
         return;
     }
 
-    /* if we're in the midst of processing something, don't die. */        
+    /* if we're in the midst of processing something, don't die. */
     if (dsi->tickle >= AFPobj->options.timeout) {
         LOG(log_error, logtype_afpd, "afp_alarm: child timed out, entering disconnected state");
         if (dsi_disconnect(dsi) != 0)
@@ -335,10 +324,10 @@ static void alarm_handler(int sig _U_)
         return;
     }
 
-    if ((err = pollvoltime(AFPobj)) == 0)
+    if ((err = pollvoltime(AFPobj)) == 0) {
         LOG(log_debug, logtype_afpd, "afp_alarm: sending DSI tickle");
-
-    err = dsi_tickle(AFPobj->handle);
+        err = dsi_tickle(AFPobj->dsi);
+    }
     if (err <= 0) {
         if (geteuid() == 0) {
             LOG(log_note, logtype_afpd, "afp_alarm: unauthenticated user, connection problem");
@@ -350,9 +339,14 @@ static void alarm_handler(int sig _U_)
     }
 }
 
-/* ----------------- 
+static void child_handler(int sig _U_)
+{
+    wait(NULL);
+}
+
+/* -----------------
    if dsi->in_write is set attention, tickle (and close?) msg
-   aren't sent. We don't care about tickle 
+   aren't sent. We don't care about tickle
 */
 static void pending_request(DSI *dsi)
 {
@@ -362,21 +356,21 @@ static void pending_request(DSI *dsi)
     if (dsi->msg_request) {
         if (dsi->msg_request == 2) {
             /* didn't send it in signal handler */
-            dsi_attention(AFPobj->handle, AFPATTN_MESG | AFPATTN_TIME(5));
+            dsi_attention(AFPobj->dsi, AFPATTN_MESG | AFPATTN_TIME(5));
         }
         dsi->msg_request = 0;
         readmessage(AFPobj);
     }
     if (dsi->down_request) {
         dsi->down_request = 0;
-        dsi_attention(AFPobj->handle, AFPATTN_SHUTDOWN | AFPATTN_NORECONNECT |
+        dsi_attention(AFPobj->dsi, AFPATTN_SHUTDOWN | AFPATTN_NORECONNECT |
                   AFPATTN_MESG | AFPATTN_TIME(5));
     }
 }
 
 void afp_over_dsi_sighandlers(AFPObj *obj)
 {
-    DSI *dsi = (DSI *) obj->handle;
+    DSI *dsi = (DSI *) obj->dsi;
     struct sigaction action;
 
     memset(&action, 0, sizeof(action));
@@ -405,7 +399,7 @@ void afp_over_dsi_sighandlers(AFPObj *obj)
     }
 
     /* install SIGQUIT */
-    action.sa_handler = ipc_reconnect_handler;
+    action.sa_handler = afp_dsi_die;
     if ( sigaction(SIGQUIT, &action, NULL ) < 0 ) {
         LOG(log_error, logtype_afpd, "afp_over_dsi: sigaction: %s", strerror(errno) );
         afp_dsi_die(EXITERR_SYS);
@@ -441,22 +435,33 @@ void afp_over_dsi_sighandlers(AFPObj *obj)
         afp_dsi_die(EXITERR_SYS);
     }
 #endif /* DEBUGGING */
+
+    /*  SIGCHLD */
+    action.sa_handler = child_handler;
+#ifdef SA_NOCLDWAIT
+/* this enhancement simplifies things for Solaris, it also improves performance */
+    action.sa_flags |= SA_NOCLDWAIT;
+#endif
+    if (sigaction(SIGCHLD, &action, NULL) < 0 ) {
+        LOG(log_error, logtype_afpd, "afp_over_dsi: sigaction: %s", strerror(errno) );
+        afp_dsi_die(EXITERR_SYS);
+    }
 }
 
 /* -------------------------------------------
- afp over dsi. this never returns. 
+ afp over dsi. this never returns.
 */
 void afp_over_dsi(AFPObj *obj)
 {
-    DSI *dsi = (DSI *) obj->handle;
+    DSI *dsi = (DSI *) obj->dsi;
     int rc_idx;
-    u_int32_t err, cmd;
-    u_int8_t function;
+    uint32_t err, cmd;
+    uint8_t function;
 
     AFPobj = obj;
     obj->exit = afp_dsi_die;
-    obj->reply = (int (*)()) dsi_cmdreply;
-    obj->attention = (int (*)(void *, AFPUserBytes)) dsi_attention;
+    obj->reply = (int (*)(void *, int))dsi_cmdreply;
+    obj->attention = (int (*)(void *, AFPUserBytes))dsi_attention;
     dsi->tickle = 0;
 
     afp_over_dsi_sighandlers(obj);
@@ -488,6 +493,8 @@ void afp_over_dsi(AFPObj *obj)
     int flag = 1;
     setsockopt(dsi->socket, SOL_TCP, TCP_NODELAY, &flag, sizeof(flag));
 
+    ipc_child_state(obj, DSI_RUNNING);
+
     /* get stuck here until the end */
     while (1) {
         if (sigsetjmp(recon_jmp, 1) != 0)
@@ -512,15 +519,6 @@ void afp_over_dsi(AFPObj *obj)
                 exit(0);
             }
 
-#if 0
-            /*  got ECONNRESET in read from client => exit*/
-            if (dsi->flags & DSI_GOT_ECONNRESET) {
-                LOG(log_note, logtype_afpd, "afp_over_dsi: client connection reset");
-                afp_dsi_close(obj);
-                exit(0);
-            }
-#endif
-
             if (dsi->flags & DSI_RECONINPROG) {
                 LOG(log_note, logtype_afpd, "afp_over_dsi: failed reconnect");
                 afp_dsi_close(obj);
@@ -531,8 +529,11 @@ void afp_over_dsi(AFPObj *obj)
             if (dsi_disconnect(dsi) != 0)
                 afp_dsi_die(EXITERR_CLNT);
 
+            ipc_child_state(obj, DSI_DISCONNECTED);
+
             while (dsi->flags & DSI_DISCONNECTED)
                 pause(); /* gets interrupted by SIGALARM or SIGURG tickle */
+            ipc_child_state(obj, DSI_RUNNING);
             continue; /* continue receiving until disconnect timer expires
                        * or a primary reconnect succeeds  */
         }
@@ -541,11 +542,12 @@ void afp_over_dsi(AFPObj *obj)
             LOG(log_debug, logtype_afpd, "afp_over_dsi: got data, ending normal sleep");
             dsi->flags &= ~DSI_SLEEPING;
             dsi->tickle = 0;
+            ipc_child_state(obj, DSI_RUNNING);
         }
 
         if (reload_request) {
             reload_request = 0;
-            load_volumes(AFPobj);
+            load_volumes(AFPobj, LV_FORCE);
         }
 
         /* The first SIGINT enables debugging, the next restores the config */
@@ -558,15 +560,15 @@ void afp_over_dsi(AFPObj *obj)
 
             if (debugging) {
                 if (obj->options.logconfig)
-                    setuplog(obj->options.logconfig);
+                    setuplog(obj->options.logconfig, obj->options.logfile);
                 else
-                    setuplog("default log_note");
+                    setuplog("default:note", NULL);
                 debugging = 0;
             } else {
                 char logstr[50];
                 debugging = 1;
-                sprintf(logstr, "default log_maxdebug /tmp/afpd.%u.XXXXXX", getpid());
-                setuplog(logstr);
+                sprintf(logstr, "/tmp/afpd.%u.XXXXXX", getpid());
+                setuplog("default:maxdebug", logstr);
             }
         }
 
@@ -591,6 +593,15 @@ void afp_over_dsi(AFPObj *obj)
             break;
 
         case DSIFUNC_CMD:
+#ifdef AFS
+            if ( writtenfork ) {
+                if ( flushfork( writtenfork ) < 0 ) {
+                    LOG(log_error, logtype_afpd, "main flushfork: %s", strerror(errno) );
+                }
+                writtenfork = NULL;
+            }
+#endif /* AFS */
+
             function = (u_char) dsi->commands[0];
 
             /* AFP replay cache */
@@ -612,10 +623,12 @@ void afp_over_dsi(AFPObj *obj)
 
                     LOG(log_debug, logtype_afpd, "<== Start AFP command: %s", AfpNum2name(function));
 
+                    AFP_AFPFUNC_START(function, (char *)AfpNum2name(function));
                     err = (*afp_switch[function])(obj,
-                                                  (char *)&dsi->commands, dsi->cmdlen,
+                                                  (char *)dsi->commands, dsi->cmdlen,
                                                   (char *)&dsi->data, &dsi->datalen);
 
+                    AFP_AFPFUNC_DONE(function, (char *)AfpNum2name(function));
                     LOG(log_debug, logtype_afpd, "==> Finished AFP command: %s -> %s",
                         AfpNum2name(function), AfpErr2name(err));
 
@@ -628,7 +641,7 @@ void afp_over_dsi(AFPObj *obj)
                     replaycache[rc_idx].AFPcommand = function;
                     replaycache[rc_idx].result = err;
                 } else {
-                    LOG(log_error, logtype_afpd, "bad function %X", function);
+                    LOG(log_maxdebug, logtype_afpd, "bad function %X", function);
                     dsi->datalen = 0;
                     err = AFPERR_NOOP;
                 }
@@ -653,10 +666,14 @@ void afp_over_dsi(AFPObj *obj)
 
                 LOG(log_debug, logtype_afpd, "<== Start AFP command: %s", AfpNum2name(function));
 
+                AFP_AFPFUNC_START(function, (char *)AfpNum2name(function));
+
                 err = (*afp_switch[function])(obj,
-                                              (char *)&dsi->commands, dsi->cmdlen,
+                                              (char *)dsi->commands, dsi->cmdlen,
                                               (char *)&dsi->data, &dsi->datalen);
 
+                AFP_AFPFUNC_DONE(function, (char *)AfpNum2name(function));
+
                 LOG(log_debug, logtype_afpd, "==> Finished AFP command: %s -> %s",
                     AfpNum2name(function), AfpErr2name(err));
 
