@@ -1,7 +1,3 @@
-/*
- * $Id: cnid_tdb_get.c,v 1.1.2.1.2.2 2005/09/27 10:40:41 didg Exp $
- */
-
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -11,14 +7,14 @@
 #include "cnid_tdb.h"
 
 /* Return CNID for a given did/name. */
-cnid_t cnid_tdb_get(struct _cnid_db *cdb, const cnid_t did, char *name, const size_t len)
+cnid_t cnid_tdb_get(struct _cnid_db *cdb, cnid_t did, const char *name, size_t len)
 {
-    char start[TDB_DID_LEN + MAXPATHLEN + 1], *buf;
+    char start[CNID_DID_LEN + MAXPATHLEN + 1], *buf;
     struct _cnid_tdb_private *db;
     TDB_DATA key, data;
     cnid_t id;
 
-    if (!cdb || !(db = cdb->_private) || (len > MAXPATHLEN)) {
+    if (!cdb || !(db = cdb->cnid_db_private) || (len > MAXPATHLEN)) {
         return 0;
     }
 
@@ -30,8 +26,8 @@ cnid_t cnid_tdb_get(struct _cnid_db *cdb, const cnid_t did, char *name, const si
     buf += sizeof(did);
     memcpy(buf, name, len);
     *(buf + len) = '\0'; /* Make it a C-string. */
-    key.dptr = start;
-    key.dsize = TDB_DID_LEN + len + 1;
+    key.dptr = (unsigned char *)start;
+    key.dsize = CNID_DID_LEN + len + 1;
     data = tdb_fetch(db->tdb_didname, key);
     if (!data.dptr)
         return 0;
